@@ -34,7 +34,7 @@ from clam.external.poster.streaminghttp import register_openers, StreamingHTTPHa
 import clam.common.status
 import clam.common.parameters
 import clam.common.formats
-from clam.common.data import CLAMData, CLAMFile, CLAMInputFile, CLAMOutputFile, CLAMMetaData, InputTemplate, OutputTemplate, VERSION as DATAAPIVERSION
+from clam.common.data import CLAMData, CLAMFile, CLAMInputFile, CLAMOutputFile, CLAMMetaData, InputTemplate, OutputTemplate, VERSION as DATAAPIVERSION, BadRequest, NotFound, PermissionDenied, ServerError, AuthRequired,NoConnection, UploadError, ParameterError, TimeOut, processhttpcode
 from clam.common.util import RequestWithMethod
 
 VERSION = '0.9.4'
@@ -45,70 +45,6 @@ if VERSION != DATAAPIVERSION:
 register_openers()
 
 
-class BadRequest(Exception):
-     def __init__(self):
-        pass
-     def __str__(self):
-        return "Bad Request"
-
-class NotFound(Exception):
-    """Raised on 404 - Not Found Errors"""
-    def __init__(self, msg=""):
-        self.msg = msg
-    def __str__(self):
-        return "Not Found: " +  self.msg
-
-class PermissionDenied(Exception):
-    """Raised on 403 - Permission Denied Errors (but only if no CLAM XML response is provided)"""
-    def __init__(self, msg = ""):
-        self.msg = msg
-    def __str__(self):
-        if isinstance(clam.common.data,CLAMData):
-            return "Permission Denied"
-        else:
-            return "Permission Denied: " + self.msg
-
-class ServerError(Exception):
-    """Raised on 500 - Internal Server Error. Indicates that something went wrong on the server side."""
-    def __init__(self, msg = ""):
-        self.msg = msg
-    def __str__(self):
-        return "Server Error: " + self.msg
-
-class AuthRequired(Exception):
-    """Raised on 401 - Authentication Required error. Service requires authentication, pass user credentials in CLAMClient constructor."""
-    def __init__(self, msg = ""):
-        self.msg = msg
-    def __str__(self):
-        return "Authorization Required: " + self.msg
-
-class NoConnection(Exception):
-    def __init__(self):
-        pass
-    def __str__(self):
-        return "Can't establish a connection with the server"
-
-
-class UploadError(Exception):
-    def __init__(self, msg = ""):
-        self.msg = msg
-    def __str__(self):
-        return "Error during Upload: " + self.msg
-
-class ParameterError(Exception):
-    """Raised on Parameter Errors, i.e. when a parameter does not validate, is missing, or is otherwise set incorrectly."""
-    def __init__(self, msg = ""):
-        self.msg = msg
-    def __str__(self):
-        return "Error setting parameter: " + self.msg
-
-class TimeOut(Exception):
-    def __init__(self):
-        pass
-    def __str__(self):
-        return "Connection with server timed-out"
-
-
 class CLAMClient:
     def __init__(self, url, user=None, password=None):
         """Initialise the CLAM client (does not actually connect yet)
@@ -156,7 +92,7 @@ class CLAMClient:
 
 
     def request(self, url='', method = 'GET', data = None):
-        """Issue a HTTP request and parse CLAM XML response, this is a low-level function called by all of the higher-level communicaton methods in this class, use them instead"""
+        """Issue a HTTP request and parse CLAM XML response, this is a low-level function called by all of the higher-level communicaton methods in this class, use those instead"""
 
         if self.authenticated: self.initauth()
         if data:
@@ -199,33 +135,8 @@ class CLAMClient:
         return self._parse(response.read())
 
 
-    def _request_httplib2(self, url, method = 'GET', data = None): #old, obsolete
-        """Returns a CLAMData object if a proper CLAM XML response is received. Otherwise: returns True, on success, False on failure.  Raises an Exception on most HTTP errors! For internal use"""
-
-        try:
-            if data:
-                response, content = self.http.request(self.url + url, method, data)
-            else:
-                response, content = self.http.request(self.url + url, method)
-        except:
-            raise NoConnection()
-        try:
-            return self._parse(response, content)
-        except TimeOut:
-            #Do it again once
-            try:
-                if data:
-                    response, content = self.http.request(self.url + url, method, data)
-                else:
-                    response, content = self.http.request(self.url + url, method)
-                return self._parse(response, content)
-            except:
-                raise
-        except:
-            raise
-
     def _parse(self, content):
-        """Parses CLAM XML data. For internal use."""
+        """Parses CLAM XML data and returns a ``CLAMData`` object. For internal use. Raises `ParameterError` exception on parameter errors."""
         if content.find('<clam') != -1:
             data = CLAMData(content,self)
             if data.errors:
@@ -236,44 +147,12 @@ class CLAMClient:
         else:
             return True
 
-    def _parse_old(self, response, content):
-        if content.find('<clam') != -1:
-            data = CLAMData(content,self)
-            if data.errors:
-                error = data.parametererror()
-                if error:
-                    raise ParameterError(error)
-        else:
-            data = False
-
-        if response['status'] == '200' or response['status'] == '201' or response['status'] == '202':
-            if not data: data = True
-        elif response['status'] == '400':
-            raise BadRequest()
-        elif response['status'] == '401':
-            raise AuthRequired()
-        elif response['status'] == '403' and data:
-            raise PermissionDenied(data)
-        elif response['status'] == '404' and data:
-            raise NotFound(content)
-        elif response['status'] == '500':
-            raise ServerError(data)
-        elif response['status'] == '408':
-            raise TimeOut()
-        else:
-            raise Exception("Server returned HTTP response " + response['status'])
-
-        return data
-
-
-
-
     def index(self):
-        """Get index of projects. Returns CLAMData instance. Use CLAMData.projects for the index of projects."""
+        """Get index of projects. Returns a ``CLAMData`` instance. Use CLAMData.projects for the index of projects."""
         return self.request('')
 
     def get(self, project):
-        """Query the project status. Returns CLAMData instance."""
+        """Query the project status. Returns a ``CLAMData`` instance or raises an exception according to the returned HTTP Status code"""
         try:
             data = self.request(project + '/')
         except:
@@ -407,11 +286,11 @@ class CLAMClient:
 
 
     def addinputfile(self, project, inputtemplate, sourcefile, **kwargs):
-        """Add/upload an input file to the CLAM service.
+        """Add/upload an input file to the CLAM service. Supports proper file upload streaming.
 
         project - the ID of the project you want to add the file to.
         inputtemplate - The input template you want to use to add this file (InputTemplate instance)
-        sourcefile - The file you want to add: either an instance of 'file' or a string containing a filename
+        sourcefile - The file you want to add: string containing a filename (or instance of ``file``)
 
         Keyword arguments (optional but recommended!):
             * ``filename`` - the filename on the server (will be same as sourcefile if not specified)
@@ -451,7 +330,7 @@ class CLAMClient:
                 assert isinstance(value, CLAMMetaData)
                 data['metadata'] =  value.xml()
             elif key == 'metafile':
-                data['metafile'] = open(value,'r')
+                data['metafile'] = (pycurl.FORM_FILE, value)  #open(value,'r')
             else:
                 data[key] = value
 
@@ -502,21 +381,8 @@ class CLAMClient:
         c.setopt(c.USERPWD, self.user + ':' + self.password)
         c.setopt(c.WRITEFUNCTION, buf.write)
         c.perform()
-        code = c.getinfo(c.HTTP_CODE)
-        if code == 200:
-            xml = buf.getvalue()
-        elif code == 400:
-            raise BadRequest()
-        elif code == 401:
-            raise AuthRequired()
-        elif code == 403:
-            raise PermissionDenied()
-        elif code == 404:
-            raise NotFound()
-        elif code == 500:
-            raise ServerError()
-        else:
-            raise UploadError()
+        code = processhttpcode(c.getinfo(c.HTTP_CODE),[403]) #raises exception when not successful
+        xml = buf.getvalue()
         c.close()
 
         try:
@@ -525,8 +391,9 @@ class CLAMClient:
             raise
 
 
+
     def addinput(self, project, inputtemplate, contents, **kwargs):
-        """Add an input file to the CLAM service. Explictly providing the contents as a string
+        """Add an input file to the CLAM service. Explictly providing the contents as a string. This is not suitable for large files as the contents are kept in memory! Use ``addinputfile()`` instead for large files.
 
         project - the ID of the project you want to add the file to.
         inputtemplate - The input template you want to use to add this file (InputTemplate instance)
@@ -596,3 +463,8 @@ class CLAMClient:
         """Alias for ``addinputfile()``"""
         return self.addinputfile(project, inputtemplate,sourcefile, **kwargs)
 
+    def download(self, project, filename, targetfilename, loadmetadata=False):
+        """Download an output file"""
+        f = CLAMOutputFile(self.url + project,  filename, loadmetadata, self)
+        f.copy(targetfilename)
+
